@@ -16,7 +16,7 @@ static const char *hs21_tiny_qp_rw(void) {
   ScsInfo info = {0};
   scs_int exitflag;
   scs_float perr, derr;
-  scs_int success;
+  scs_int success, read_status;
   const char *fail;
 
   /* data */
@@ -84,7 +84,12 @@ static const char *hs21_tiny_qp_rw(void) {
   scs_free(stgs);
   scs_free(d);
 
-  SCS(read_data)("hs21_tiny_qp", &d, &k, &stgs);
+  read_status = SCS(read_data)("hs21_tiny_qp", &d, &k, &stgs);
+
+  if (read_status < 0) {
+    return "Data read failure, exit.\n";
+  }
+
   stgs->max_iters = 1000;
   /* solve with read data */
   exitflag = scs(d, k, stgs, sol, &info);
